typedef unsigned long			Bit32u;
typedef unsigned int			Bit16u;
typedef unsigned char			Bit8u;

typedef signed int				Bit16s;

#include <dos.h>
#include <bios.h>

typedef struct REGPACK REGSTRUCT;

static Bit16u readBiosDataArea(Bit16u offset)
{
	Bit16u retval;
	asm push es;
	asm mov ax, 040h;
	asm mov es, ax;
	asm mov bx,offset;
	asm mov dx, es:bx;
	asm mov retval, dx;
	asm pop es;
	
	return retval;
}

static Bit8u lastChar;

static Bit8u checkChar() {
	Bit16u temp = readBiosDataArea(0x1a);
	if(temp != readBiosDataArea(0x1c)) {
		lastChar = getch();
		return lastChar;
	}
	return 0;
}


static void writeSegmentB(Bit16u offset,Bit8u data)
{
	asm push es;
	asm mov ax, 0b000h;
	asm mov es, ax;
	asm mov al, data;
	asm mov bx,offset;
	asm mov es:bx, al;
	asm pop es;
}

static void writeSegmentA(Bit16u offset,Bit8u data)
{
	asm push es;
	asm mov ax, 0a000h;
	asm mov es, ax;
	asm mov al, data;
	asm mov bx,offset;
	asm mov es:bx, al;
	asm pop es;
}

Bit16u setStdMode(Bit8u mode)
{
	REGSTRUCT regs;
	regs.r_ax=mode;
	regs.r_bx=0; 
	regs.r_cx=0;
	regs.r_dx=0;
	intr(0x10, &regs);
	if((readBiosDataArea(0x49)&0xFF)!=mode) return 0;
	return 1;
}

Bit8u setVesaMode(Bit16u mode)
{
	REGSTRUCT regs;
	regs.r_ax=0x4f02;
	regs.r_bx=mode; 
	regs.r_cx=0;
	regs.r_dx=0;
	intr(0x10, &regs);
	if((regs.r_ax>>8)==1) return 0;
	return 1;
}

void setDisplayStart(Bit16u start) {
	outport(0x3d4, 0xd | (start<<8));		/* lower */
	outport(0x3d4, 0xc | (start&0xff00));	/* upper */
}

void setSplitline(Bit16u split) {
	Bit16u temp;
	outport(0x3d4, 0x18 | ((split&0xff)<<8));

	outportb(0x3d4, 0x7);
	temp = inport(0x3d5);
	temp &= ~0x10;
	if(split&0x100) temp |= 0x10;
	outport(0x3d4, 0x7 | temp<<8);

	outportb(0x3d4, 0x9);
	temp = inportb(0x3d5);
	temp &= ~0x40;
	if(split&0x200) temp |= 0x40;
	outport(0x3d4, 0x9 | temp<<8);
}

void setPanningCompat(Bit8u value) {
	Bit8u temp;
	inportb(0x3da);
	outportb(0x3c0,0x10);
	temp = inportb(0x3c1);
	inportb(0x3da);
	outportb(0x3c0,0x10);
	temp &= ~0x20;
	if(!value) temp |= 0x20;
	outportb(0x3c0,temp);

	outportb(0x3c0,0x20);
}

Bit8u getScanLength() {
	outportb(0x3d4, 0x13);
	return inportb(0x3d5);
}

void setScanLength(Bit16u value) {
	outport(0x3d4, 0x13 | (value<<8));
}

void waitForDisplayStart() {
	/* wait for retrace */
	while(!(inportb(0x3da)&0x8));

	/* wait for blanking off */
	while(inportb(0x3da)&0x1);

}

void waitTillVretStart() {
	/* wait for retrace */
	while(!(inportb(0x3da)&0x8));
}

void waitTillVretEnd() {
	/* wait for retrace */
	while(!(inportb(0x3da)&0x8));
	/* wait for retrace off */
	while(inportb(0x3da)&0x8);
}

void waitTillDisplayStart() {
	/* wait for blanking off */
	while(inportb(0x3da)&0x1);
}

void skipLines(Bit8u number) {
	while(number) {
		while(!(inportb(0x3da)&0x1));
		while(inportb(0x3da)&0x1);
		number--;
	}
}

void skipFrames(Bit16u number) {
	while(number) {
		waitForDisplayStart();
		number--;
	}
}

void setPanning(Bit8u val) {
	inportb(0x3da);
	outportb(0x3c0,0x13);
	outportb(0x3c0,val);
	inportb(0x3da);
	outportb(0x3c0,0x20);
}

void customPrint(Bit8u page, Bit8u rows, Bit16u cols, Bit8u x, Bit8u y,
				 const Bit8u* string, Bit8u attrib) {
	Bit16u addr=rows*cols*2*page + cols*y*2+x*2;
	Bit8u index=0;
	while(string[index]!=0) {
		writeSegmentB(addr + 2*index + 0x8000,string[index]);
		writeSegmentB(addr + 2*index + 0x8000 + 1,attrib);
		index++;
	}
	
}

/****************** test pattern drawer *****************/
void testpattern_text() {
	Bit16u i;
	for(i=0; i < 32768; i++) {
		Bit8u character = (i>>1)&0xff;
		Bit8u attr = (i>>9)&0xff;
		
		if(i&1) writeSegmentB(i+0x8000,attr);
		else writeSegmentB(i+0x8000,character);
	}
}

void testpattern_graphics() {
	Bit16u i;
	for(i=0; i < 65534; i++) {
		Bit8u character = (i>>1)&0xff;
		Bit8u attr = (i>>9)&0xff;
		
		if(i&1) writeSegmentA(i,attr+1);
		else writeSegmentA(i,character);
		if(i&1) writeSegmentB(i,attr+1);
		else writeSegmentB(i,character);
	}
}



const char* splittext = "Should not change when reaching top... ";
const char* maintext   = "Test for VGA text mode display start timing - should neither move nor flicker...";
const char* pantimtext = "Test for VGA text mode panning timing - should neither move nor flicker...      ";
const char* textwidth = "<- 0 --------------------------------------------------------------------- 80 -><- 81 ------------------------------------------------------------------- 160 -><- 161 ---";
const char* panningtext = "Vertical scrolling test - should be smooooooth... ";
const char* panningtext2 = "Horizontal panning...         vgaonly: watch the background -> ";




void runTextTest(Bit8u mode, Bit8u initialcols, Bit8u rows,
				 Bit16u splitmax, Bit8u charwidth) {
	Bit16s i;
	Bit16u temp;
	Bit16u virtual_width;
	Bit8u buffer[20];

	sprintf(buffer,"Video mode: %x",mode);
	if(!setStdMode(mode)) goto TEXT_DONE;
	waitTillVretStart();
	virtual_width = getScanLength() * 2 * 2;
		if(virtual_width>511) virtual_width=511;

	setScanLength(virtual_width/2);


	setDisplayStart(virtual_width * rows);

	testpattern_text();

	customPrint(0, rows, virtual_width, 0, 0, splittext, 0xff);
	customPrint(0, rows, virtual_width, 0, 1, buffer, 0xff);
	customPrint(1, rows, virtual_width, 0, 0, maintext, 0xff);
	customPrint(0, rows, virtual_width, 0, 2, "ESC to quit, other key to skip this mode", 0xff);
	/*customPrint(0, rows, virtual_width, 0, 3, textwidth, 0xff);*/

	/* close */
	for(i=splitmax; i > -1; i--) {
		waitTillVretStart();
		outport(0x3d4, 0xc | (i&0xf));/* = setDisplayStart(i&0xf); /* wrong */
		setSplitline(i);
		waitTillDisplayStart();
		skipLines(10);
		setDisplayStart(initialcols*2*rows); /* right */
		if(checkChar())goto TEXT_DONE;
	}
		
	customPrint(1, rows, initialcols*2, 0, 0, pantimtext, 0xff);
	skipFrames(70);

	/* open */
	for(i=0; i <=splitmax; i++) {
		if(checkChar())goto TEXT_DONE;
		waitTillVretStart();

		setPanning((charwidth==9)?8:0); /* right panning */
		waitTillVretEnd();

		setPanning(i&0xf); /* wrong panning */

		setSplitline(i);
		waitTillDisplayStart();
		skipLines(150);
		while(!(inportb(0x3da)&0x1)); /* wait for blanking */
	}
	setPanning((charwidth==9)?8:0); /* right panning */
	
	skipFrames(70);

	testpattern_text();
	
	customPrint(1, rows, initialcols*2, 0, 0, panningtext, 0xff);

	/* scroll up */
	for(i=rows*16; i >-1; i--) {
		
		skipFrames(1);
		outport(0x3d4, 0x8 | ((i%16)<<8));
		setDisplayStart(virtual_width*(i/16));
		
		if(checkChar())goto TEXT_DONE;
	}

	setDisplayStart(0);

	/* split-panning off */
	setPanningCompat(0);

	customPrint(0, rows, initialcols*2, 0, 0, panningtext2, 0xff);

	/* scroll side */
	for(i=0; i <(initialcols*9)-1; i++) {
		
		skipFrames(1);
		if(i<(splitmax/3)) setSplitline(splitmax-i);
		temp=(i%9)-1;
		if(temp>8) temp=8;
		setPanning(temp);

		outport(0x3d4, 0x8 | ( ((i/9)&0x3) <<(8+5)));
		setDisplayStart((i/9)&~0x3);

		if(checkChar())goto TEXT_DONE;


	}
	skipFrames(70);
TEXT_DONE:
	setStdMode(3);
}

void runGraphicsTest(Bit8u mode, Bit16u bitperpix, Bit16u pixperaddr,
				 Bit16u width, Bit16u height, Bit16u splitmax) {
	Bit16s i;
	Bit16u temp;
	Bit16u virtual_width;
	Bit16u currentStart;
	Bit8u offset;
	Bit8u buffer[20];


	/* current line of display start */
	currentStart = (width/pixperaddr)*(height/2);



	sprintf(buffer,"Video mode: %x",mode);
	if(!setStdMode(mode)) goto GRAPH_DONE;
	waitTillVretStart();
	/*virtual_width = width * 2;*/
	virtual_width = width;
	
	/*offset=(virtual_width)/(pixperaddr*2); /* word mode?! */
	/*setScanLength(offset);*/


	/*setDisplayStart((virtual_width/pixperaddr)*height);*/
	setDisplayStart(currentStart);
	
	testpattern_graphics();


	puts(splittext);
	sprintf(buffer,"Video mode: %x ",mode);
	puts(buffer);
	puts("Upper part should neither move nor \nflicker ");

	/* close */
	for(i=splitmax; i > -1; i--) {
		waitTillVretStart();
		outport(0x3d4, 0xc | (i&0xf));/* = setDisplayStart(i&0xf); /* wrong */
		setSplitline(i);
		waitTillDisplayStart();
		skipLines(10);
		setDisplayStart(currentStart); /* right */
		if(checkChar())goto GRAPH_DONE;
	}
		
	skipFrames(70);

	/* open */
	for(i=0; i <=splitmax; i++) {

		waitTillVretStart();

		setPanning(0); /* right panning */
		waitTillVretEnd();

		setPanning(i&0xf); /* wrong panning */

		setSplitline(i);
		waitTillDisplayStart();
		skipLines(150);
		if(checkChar())goto GRAPH_DONE;
		while(!(inportb(0x3da)&0x1)); /* wait for blanking */
	}
	setPanning(0); /* right panning */
	
	skipFrames(70);

	testpattern_graphics();
	
	/* scroll up */
	for(i=0; i <height; i++) {
		waitTillVretEnd();
		/*skipFrames(1);*/
		/*outport(0x3d4, 0x8 | ((i%16)<<8));*/
		setDisplayStart(currentStart-(width/pixperaddr)*i);
		if(checkChar())goto GRAPH_DONE;
	}

	skipFrames(70);

	/* split-panning off */
	setPanningCompat(0);

	setScanLength(getScanLength()*2);

	/* scroll side */
	for(i=0; i < width; i++) {
		
		skipFrames(1);
		if(i<(splitmax/3)) setSplitline(splitmax-i);

		
		if(bitperpix==4) {
			setPanning((i-1)%8);
			setDisplayStart(((i+7)>>3));
		}
		if(bitperpix==8) {
			setPanning(((i-1)%4)*2);
			setDisplayStart(((i+3)>>2));	
		} else {
			setPanning((i-1)%8);
			setDisplayStart(((i+7)>>3));
		}
		
		

		/*outport(0x3d4, 0x8 | ( ((i/9)&0x3) <<(8+5)));*/
		
		/*setDisplayStart(i/bitperpix);*/

		if(checkChar())goto GRAPH_DONE;
	}

	skipFrames(70);
GRAPH_DONE:
	setStdMode(3);
}


void main(int argc, char argv[])
{
	while(1) {

		runGraphicsTest(0x04, 2, 4, 320, 200, 400);		if(lastChar==0x1b) return;
		runGraphicsTest(0x05, 2, 4, 320, 200, 400);		if(lastChar==0x1b) return;
		runGraphicsTest(0x06, 2, 4, 640, 200, 400);		if(lastChar==0x1b) return;

		runGraphicsTest(0x0d, 4, 4, 320, 200, 400);		if(lastChar==0x1b) return;
		runGraphicsTest(0x0e, 4, 4, 640, 200, 400);		if(lastChar==0x1b) return;
		runGraphicsTest(0x10, 4, 4, 640, 350, 350);		if(lastChar==0x1b) return;
		
		runGraphicsTest(0x12, 4, 4, 640, 480, 480);		if(lastChar==0x1b) return;
		runGraphicsTest(0x13, 8, 4, 320, 200, 400);		if(lastChar==0x1b) return;
			
		
		
		runTextTest(0x00,  40, 25, 400, 9);				if(lastChar==0x1b) return;
		runTextTest(0x01,  40, 25, 400, 9);				if(lastChar==0x1b) return;
		runTextTest(0x02,  80, 25, 400, 9);				if(lastChar==0x1b) return;
		runTextTest(0x03,  80, 25, 400, 9);				if(lastChar==0x1b) return;
		/*runTextTest(0x7, 80, 25, 400, 9);				if(lastChar==0x1b) return; DOSBox locks up*/
		/*runTextTest(0x54, 132, 43, 688, 8);				if(lastChar==0x1b) return; /* S3 specific, real hw apparently doesn't support scroll up */




	}
}

